import pytest

from schocken.spiel import Einwerfen, SchockenSpiel
from schocken.spieler import Spieler
from schocken.exceptions import FalscheAktion, FalscherSpieler, NochNichtGeworfen

from schocken import wuerfel


@pytest.fixture
def spieler(n=4):
    # testspieler
    return [Spieler(f"spieler_{i+1}") for i in range(n)]


def test_uebergang_einwerfen_halbzeit(spieler):
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"

    # Übergang von Einwerfen nach Halbzeit
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    wuerfel.werfen = lambda n: (3,) * n
    runde.command_to_event(spieler[1].name, "einwerfen")
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[2].name, "einwerfen")

    wuerfel.werfen = lambda n: (4, 6, 2)
    runde.command_to_event(spieler[2].name, "wuerfeln")

    assert runde._leaf_state.name == "wuerfeln"


@pytest.fixture
def drei_spieler_eingeworfen_spieler_zwei_muss_werfen(spieler):
    # FALL: spieler 2 würfelt eindeutig das niedrigste
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[1].name, "einwerfen")
    wuerfel.werfen = lambda n: (3,) * n
    runde.command_to_event(spieler[2].name, "einwerfen")

    return runde


def test_falscher_spieler(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    wuerfel.werfen = lambda n: (3, 1, 6) * n
    runde.command_to_event(spieler[1].name, "wuerfeln")

    # Spieler 1 führt mehrere Argumente aus
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, "wuerfeln")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, "beiseite legen")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, "weiter")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, "umdrehen")


def test_falsche_aktion(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    wuerfel.werfen = lambda n: (3, 2, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")

    # Spieler 2 führt falsche Aktionen aus
    with pytest.raises(FalscheAktion):
        # keine Eins gewürfelt
        runde.command_to_event(spieler[1].name, "beiseite legen")

    with pytest.raises(FalscheAktion):
        # nur eine Sechs gewürfelt
        runde.command_to_event(spieler[1].name, "umdrehen")


def test_sechsen_umdrehen(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    # Spieler 2 würfelt zweimal zwei Sechsen und dreht diese um
    wuerfel.werfen = lambda n: (3, 6, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "umdrehen")

    # Mehrmaliges umdrehen derselben Sechsen ist nicht erlaubt
    with pytest.raises(FalscheAktion):
        runde.command_to_event(spieler[1].name, "umdrehen")

    wuerfel.werfen = lambda n: (6, 6,)
    runde.command_to_event(spieler[1].name, "wuerfeln")

    assert runde.halbzeit_erste.spieler_liste[0].augen == (1, 6, 6)

    runde.command_to_event(spieler[1].name, "umdrehen")
    wuerfel.werfen = lambda n: (2,)
    runde.command_to_event(spieler[1].name, "wuerfeln")

    assert runde.halbzeit_erste.spieler_liste[0].augen == (1, 1, 2)

    # Spieler 2 will noch mehr machen
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, "weiter")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, "wuerfeln")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, "beiseite legen")

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, "umdrehen")


def test_einsen_beiseite_legen(
    spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen
):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    # Spieler 2 würfelt zweimal zwei Einsen und legt diese nacheinander zur Seite
    wuerfel.werfen = lambda n: (1, 2, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "beiseite legen")
    # zweimal zur Seite legen derselben Eins ist nicht!
    with pytest.raises(FalscheAktion):
        runde.command_to_event(spieler[1].name, "beiseite legen")

    wuerfel.werfen = lambda n: (1, 2,)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "beiseite legen")

    wuerfel.werfen = lambda n: (6,)
    runde.command_to_event(spieler[1].name, "wuerfeln")

    assert runde.halbzeit_erste.spieler_liste[0].augen == (1, 1, 6)


def test_einsen_und_sechsen(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    # Kombination aus Einsen beiseite legen und Sechsen umdrehen
    wuerfel.werfen = lambda n: (5, 6, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "umdrehen")

    wuerfel.werfen = lambda n: (2, 1,)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "beiseite legen")

    wuerfel.werfen = lambda n: (3,)
    runde.command_to_event(spieler[1].name, "wuerfeln")

    assert runde.halbzeit_erste.spieler_liste[0].augen == (1, 1, 3)

    assert runde.halbzeit_erste.aktiver_spieler.name == "spieler_3"


def test_nachgeworfen(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    # Alle Spieler würfeln dasselbe, letzter Spieler verliert die Runde
    wuerfel.werfen = lambda n: (5, 6, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "weiter")
    runde.command_to_event(spieler[2].name, "wuerfeln")
    runde.command_to_event(spieler[0].name, "wuerfeln")

    runde.halbzeit_erste.spieler_liste[2].deckel == 1


def test_uebergang_erste_halbzeit_zweite_halbzeit(
    spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen
):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    # Herrenwurf verliert nicht ;)
    wuerfel.werfen = lambda n: (5, 5, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "weiter")
    wuerfel.werfen = lambda n: (1, 1, 3)
    runde.command_to_event(spieler[2].name, "wuerfeln")
    wuerfel.werfen = lambda n: (1, 1, 1)
    runde.command_to_event(spieler[0].name, "wuerfeln")

    # Erste Halbzeit ist vorbei und befindet sich im state_stack der root_machine
    assert runde.state_stack.deque[1].name == "Halbzeit"
    # Startspieler ist Spieler 2
    assert runde.halbzeit_zweite.spieler_liste[0].name == spieler[1].name
    # Spielerreihenfolge ist nun Spieler 2, Spieler 3, Spieler 1
    assert runde.halbzeit_zweite.spieler_liste[1].name == spieler[2].name
    assert runde.halbzeit_zweite.spieler_liste[2].name == spieler[0].name
    # Alle Spieler haben wieder Null Deckel
    assert runde.halbzeit_zweite.spieler_liste[0].deckel == 0
    assert runde.halbzeit_zweite.spieler_liste[1].deckel == 0
    assert runde.halbzeit_zweite.spieler_liste[2].deckel == 0


@pytest.fixture
def erste_halbzeit_beendet(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen

    wuerfel.werfen = lambda n: (5, 5, 6)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "weiter")
    wuerfel.werfen = lambda n: (1, 1, 3)
    runde.command_to_event(spieler[2].name, "wuerfeln")
    wuerfel.werfen = lambda n: (1, 1, 1)
    runde.command_to_event(spieler[0].name, "wuerfeln")

    return runde


def test_uebergang_zweite_halbzeit_finale_mit_zwei_spielern(
    spieler, erste_halbzeit_beendet
):
    runde = erste_halbzeit_beendet

    # Erste Halbzeit verlor Spieler 2, diese wird Spieler 3 verlieren
    wuerfel.werfen = lambda n: (1, 1, 1)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "weiter")
    wuerfel.werfen = lambda n: (2, 2, 3)
    runde.command_to_event(spieler[2].name, "wuerfeln")
    wuerfel.werfen = lambda n: (1, 2, 1)
    runde.command_to_event(spieler[0].name, "wuerfeln")

    # Zweite Halbzeit ist vorbei und befindet sich im state_stack der root_machine
    assert runde.state_stack.deque[2].name == "Halbzeit"
    # Startspieler ist Spieler 1
    assert runde.finale.spieler_liste[0].name == spieler[2].name
    # Zweiter Spieler ist Spieler 2
    assert runde.finale.spieler_liste[1].name == spieler[1].name
    # Finalisten haben wieder Null Deckel
    assert runde.finale.spieler_liste[0].deckel == 0
    assert runde.finale.spieler_liste[1].deckel == 0


def test_ein_spieler_verliert_beide_halbzeiten(spieler, erste_halbzeit_beendet):
    runde = erste_halbzeit_beendet

    # Erste Halbzeit verlor Spieler 2, diese wird er ebenso verlieren
    wuerfel.werfen = lambda n: (1, 2, 2)
    runde.command_to_event(spieler[1].name, "wuerfeln")
    runde.command_to_event(spieler[1].name, "weiter")
    wuerfel.werfen = lambda n: (2, 2, 3)
    runde.command_to_event(spieler[2].name, "wuerfeln")
    wuerfel.werfen = lambda n: (1, 1, 1)
    runde.command_to_event(spieler[0].name, "wuerfeln")

    # Zweite Halbzeit ist vorbei und befindet sich im state_stack der root_machine
    assert runde.state_stack.deque[2].name == "Halbzeit"
    # root_machine sollte sich nun im state "anstoßen!" befinden
    assert runde.state.name == "anstoßen!"



import os

import discord
from discord.utils import get
from dotenv import load_dotenv

from schocken import __version__
from schocken.exceptions import FalscherServer
from schocken.bot import SchockenBot


def run_bot():
    load_dotenv()
    TOKEN = os.getenv("DISCORD_TOKEN")
    GUILD = os.getenv("DISCORD_GUILD")
    client = discord.Client()

    # when connecting
    @client.event
    async def on_ready():
        client.bot = SchockenBot(client)
        # check if server is correct:
        for guild in client.guilds:
            if guild.name != client.bot.valid_guild_name:
                raise FalscherServer("Dieser Bot darf nur ins Café A")

        ch = client.get_channel(690929770355097610)  # schocktresen
        await ch.send(f"Schocken (v{__version__}) kann jetzt losgehen. :muscle:")
        print("Success")
        return

    # when a message is read by the bot
    @client.event
    async def on_message(message):
        await client.bot.parse_input(message)
        return

    @client.event
    async def on_error(error, *args, **kwargs):
        import traceback

        ch = client.get_channel(694603857950539887)  # errorland
        trace = traceback.format_exc()
        await ch.send(f"Error: {error}({args}, {kwargs})\n```\n{trace}\n```")
        return

    client.run(TOKEN)


if __name__ == "__main__":
    run_bot()



import pytest

from schocken import wurf
from schocken.spieler import Spieler
from schocken.deckel_management import RundenDeckelManagement, SpielzeitStatus

NUM_START_DECKEL = 15


@pytest.fixture
def neue_runde() -> RundenDeckelManagement:
    spielerinnen = [Spieler(name) for name in "ABC"]
    szs = SpielzeitStatus(NUM_START_DECKEL, spielerinnen)
    rdm = RundenDeckelManagement(szs)
    return rdm


def test_ist_lust_wurf(neue_runde):
    rdm = neue_runde
    erster_wurf = rdm.wurf("A", (2, 2, 1), aus_der_hand=True)
    zweiter_wurf = rdm.wurf("A", (1, 1, 1), aus_der_hand=False)
    assert not rdm.ist_lust_wurf(erster_wurf)
    assert not rdm.ist_lust_wurf(zweiter_wurf)

    rdm.weiter()
    erster_wurf = rdm.wurf("B", (5, 5, 6), aus_der_hand=True)
    zweiter_wurf = rdm.wurf("B", (5, 5, 6), aus_der_hand=True)
    assert not rdm.ist_lust_wurf(erster_wurf)
    assert not rdm.ist_lust_wurf(zweiter_wurf)

    rdm.weiter()
    erster_wurf = rdm.wurf("C", (1, 2, 3), aus_der_hand=True)
    zweiter_wurf = rdm.wurf("C", (1, 2, 3), aus_der_hand=True)

    assert not rdm.ist_lust_wurf(erster_wurf)
    assert rdm.ist_lust_wurf(zweiter_wurf)


def test_schockout_verteilen_einen(neue_runde: RundenDeckelManagement):
    rdm = neue_runde
    rdm.wurf("A", (1, 2, 3), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("B", (1, 1, 1), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("C", (5, 5, 6), aus_der_hand=True)

    neuer_status = rdm.deckel_verteilen_restliche_spieler()
    assert neuer_status.deckel_in_topf == 0
    assert len(neuer_status.spieler) == 1
    assert neuer_status.spieler[0].deckel == NUM_START_DECKEL
    assert neuer_status.spieler[0].name == "C"


def test_schockout_verteilen_zwei(neue_runde: RundenDeckelManagement):
    rdm = neue_runde
    rdm.wurf("A", (1, 1, 1), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("B", (1, 1, 1), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("C", (5, 5, 6), aus_der_hand=True)

    neuer_status = rdm.deckel_verteilen_restliche_spieler()
    assert neuer_status.deckel_in_topf == 0
    assert len(neuer_status.spieler) == 1
    assert neuer_status.spieler[0].deckel == NUM_START_DECKEL
    assert neuer_status.spieler[0].name == "C"


def test_schockout_verteilen_alle(neue_runde: RundenDeckelManagement):
    rdm = neue_runde
    rdm.wurf("A", (1, 1, 1), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("B", (1, 1, 1), aus_der_hand=True)
    rdm.weiter()
    rdm.wurf("C", (1, 1, 1), aus_der_hand=True)

    neuer_status = rdm.deckel_verteilen_restliche_spieler()
    assert neuer_status.deckel_in_topf == 0
    assert len(neuer_status.spieler) == 1
    assert neuer_status.spieler[0].deckel == NUM_START_DECKEL
    assert neuer_status.spieler[0].name == "A"


def test_simulate_spielzeit_naive():
    A = Spieler("A")
    B = Spieler("B")

    spielzeit_status = SpielzeitStatus(15, [A, B])
    while len(spielzeit_status.spieler) > 1:
        print(f"Status bei Rundenbeginn: {spielzeit_status}")
        spielzeit_status = _runde_spielen_jule_im_dritten(spielzeit_status)
        print(f"Status bei Rundenende: {spielzeit_status}")
    print(f"{spielzeit_status.spieler[0]} hat verloren!")


def _runde_spielen_schockout_im_ersten(status: SpielzeitStatus) -> SpielzeitStatus:
    jules_augen = wurf.SonderWurf.Jule._value_
    schockout_augen = wurf.Schock.out._value_
    rdm = RundenDeckelManagement(status)

    aktueller_spieler = 0
    rdm.wurf(status.spieler[aktueller_spieler].name, jules_augen, aus_der_hand=True)
    aktueller_spieler = rdm.weiter()
    rdm.wurf(status.spieler[aktueller_spieler].name, schockout_augen, aus_der_hand=True)
    return rdm.deckel_verteilen_restliche_spieler()


def _runde_spielen_jule_im_dritten(status: SpielzeitStatus) -> SpielzeitStatus:
    jules_augen = wurf.SonderWurf.Jule._value_
    rdm = RundenDeckelManagement(status)

    aktueller_spieler = 0
    for _ in range(3):
        rdm.wurf(status.spieler[aktueller_spieler].name, jules_augen, aus_der_hand=True)
    aktueller_spieler = rdm.weiter()
    # zweiter spieler muss gewinnen, da er/sie sonst anfaengt
    # und es zu einer Endlos-Schleife kommt
    for _ in range(2):
        rdm.wurf(status.spieler[aktueller_spieler].name, jules_augen, aus_der_hand=True)
    return rdm.deckel_verteilen_restliche_spieler()



import os
import signal

from .exceptions import (
    SpielLaeuft,
    SpielLaeuftNicht,
    FalscherSpielBefehl,
    FalscheAktion,
    PermissionError,
    FalscherSpieler,
    ZuOftGeworfen,
)
from .spiel import SchockenSpiel
from . import wurf
from discord.utils import get
from copy import deepcopy
import random


class SchockenBot:
    def __init__(self, client):
        self.client = client
        # bot will never run on any other server than Café A
        self.guild = client.guilds[0]
        self.schock_channel_name = "schocktresen"
        self.valid_guild_name = "Café A"
        self.game_running = False
        self._all_member_names = [member.name for member in self.guild.members]
        self._start_game_cmd = "schocken"
        self._end_game_cmd = "beenden"
        self._restart_cmd = "neustart"
        self._wuerfel_emoji_names = dict(
            [
                (1, "wuerfel_1"),
                (2, "wuerfel_2"),
                (3, "wuerfel_3"),
                (4, "wuerfel_4"),
                (5, "wuerfel_5"),
                (6, "wuerfel_6"),
            ]
        )
        self.discord_to_game_cmd_dict = {
            "einwerfen": "einwerfen",
            "wuerfeln": "wuerfeln",
            "stechen": "stechen",
            "weiter": "weiter",
            "beiseite": "beiseite legen",
        }

        self.game_to_discord_cmd_dict = {
            v: k for k, v in self.discord_to_game_cmd_dict.items()
        }

        self._halbzeit_state_names = {
            1: "halbzeit_erste",
            2: "halbzeit_zweite",
            3: "finale",
        }

    def emoji_by_name(self, name):
        emoji = get(self.guild.emojis, name=name)
        return str(emoji)

    def name_to_member(self, name):
        member = get(self.guild.members, name=name)
        return member

    def wurf_to_emoji(self, wuerfe):
        out = " ".join(
            [self.emoji_by_name(self._wuerfel_emoji_names[wurf]) for wurf in wuerfe]
        )
        return out

    def discord_to_game_cmd(self, discord_cmd):
        try:
            game_cmd = self.discord_to_game_cmd_dict[discord_cmd]
            return game_cmd
        except KeyError:
            raise FalscherSpielBefehl

    def spieler_by_name(self, name, spielerliste):
        spieler = next(sp for sp in spielerliste if sp.name == name)
        return spieler

    def replace_names_by_mentions(self, string):
        for name in self._all_member_names:
            member = self.name_to_member(name)
            string = string.replace(name, member.mention)
        return string

    def command_in_schock_channel(self, message):
        msg_text = message.content
        channel = message.channel
        correct_channel = channel.name == self.schock_channel_name
        is_command = msg_text.startswith("!")
        is_not_restart = self._restart_cmd not in msg_text
        return correct_channel and is_command and is_not_restart

    def restart_issued(self, message):
        msg_text = message.content
        return msg_text == f"!{self._restart_cmd}"

    async def parse_input(self, message):
        # all messages from channels with read permissions are read
        msg_text = message.content
        channel = message.channel
        try:
            if self.command_in_schock_channel(message):
                command = msg_text.split("!")[-1]
                if command == self._start_game_cmd:
                    # TODO Status auf Spiel läuft setzten
                    if self.game_running:
                        raise SpielLaeuft
                    else:
                        self.game_running = True
                        self.game = SchockenSpiel()
                        msg = f"{message.author.mention} will schocken. `!einwerfen` zum mitmachen"
                        await self.print_to_channel(channel, msg)

                elif command == self._end_game_cmd:
                    # TODO Status auf Spiel beendet setzten
                    if self.game_running:
                        msg = f"{message.author.mention} hat das Spiel beendet"
                        self.game_running = False
                        await self.print_to_channel(channel, msg)
                    else:
                        raise SpielLaeuftNicht

                elif command == "ICH WILL UNREAL TOURNAMENT SPIELEN":
                    msg = "Dann geh doch"
                    await self.print_to_channel(channel, msg)
                    link = "https://tenor.com/view/unreal-tournament-kid-unreal-unreal-kid-rage-gif-16110833"
                    await self.print_to_channel(channel, link)

                else:
                    if not self.game_running:
                        raise SpielLaeuftNicht
                    # actual game
                    await self.handle_game(message)

            elif self.restart_issued(message):
                role_strs = [str(role) for role in message.author.roles]
                if "developer" not in role_strs:
                    raise PermissionError
                msg = f"👋 Bis gleich! :wave:"
                await self.print_to_channel(channel, msg)
                await self.client.logout()
                os.kill(os.getpid(), signal.SIGINT)

        except NotImplementedError:
            msg = "Das geht leider noch nicht. (Nicht implementiert)"
            msg += "\n Spiel wird beendet."
            await self.print_to_channel(channel, msg)
            self.game_running = False
            del self.game

        except PermissionError:
            msg = f"Das darfst du nicht, DU HURENSOHN! {self.emoji_by_name('king')}"
            await self.print_to_channel(channel, msg)

        except SpielLaeuftNicht:
            msg = f"Gerade läuft kein Spiel. `!{self._start_game_cmd}` zum starten"
            await self.print_to_channel(channel, msg)

        except SpielLaeuft:
            msg = f"Es läuft bereits ein Spiel, schau, ob du `!einwerfen` kannst."
            await self.print_to_channel(channel, msg)

        except FalscherSpielBefehl:
            # avail_handlers = self.game.leaf_state.handlers.keys()
            # cmds = [
            # f"`!{self.game_to_discord_cmd_dict[hdlr]}`" for hdlr in avail_handlers
            # ]
            msg = (
                "Diesen Befehl gibt es nicht."  # Versuch's mal mit einem von diesen:\n"
            )
            # msg += ", ".join(cmds)
            # msg += "\n".join(["`" + bef + "`" for bef in zulaessig])
            await self.print_to_channel(channel, msg)

        except FalscherSpieler as e:
            if str(e):
                msg = self.replace_names_by_mentions(str(e))
            else:
                msg = "Das darfst du gerade nicht (Falsche Spielerin)."
            await self.print_to_channel(channel, msg)

        except ZuOftGeworfen as e:
            if str(e):
                msg = self.replace_names_by_mentions(str(e))
            else:
                msg = "Du darfst nicht nochmal!"
            await self.print_to_channel(channel, msg)

        except FalscheAktion as e:
            if str(e):
                msg = self.replace_names_by_mentions(str(e))
            else:
                msg = "Das darfst du gerade nicht. (Falsche Aktion)"
            await self.print_to_channel(channel, msg)

    async def print_to_channel(self, channel, text):
        return await channel.send(text)

    async def handle_game(self, message):
        msg_text = message.content
        channel = message.channel
        msg_author = message.author
        command = msg_text.split("!")[-1]
        spieler_name = msg_author.name

        # freeze old game state. some properties are needed for the bot
        game_old = deepcopy(self.game)
        # run game state machine
        game_cmd = self.discord_to_game_cmd(command)
        if game_cmd not in self.game.leaf_state.handlers.keys():
            raise FalscherSpielBefehl
        self.game.command_to_event(spieler_name, game_cmd)
        state_new = self.game.state.leaf_state.name

        if state_new == "einwerfen":
            # einwerfen is the initial state, no need to check for changes
            spieler = self.spieler_by_name(
                spieler_name, self.game.einwerfen.spieler_liste
            )
            out_str = f"{message.author.mention} hat eine {self.emoji_by_name(self._wuerfel_emoji_names[spieler.augen])} geworfen."
            # stechen oder nicht?
            if self.game.einwerfen.stecher_count > 1:
                out_str += "\n" + (
                    ", ".join(
                        [
                            self.name_to_member(pl.name).mention
                            for pl in self.game.einwerfen.stecher_liste
                        ]
                    )
                    + f" haben eine {self.emoji_by_name(self._wuerfel_emoji_names[self.game.einwerfen.spieler_liste[0].augen])} geworfen.\n"
                )
                out_str += (
                    "`!stechen` um zu stechen oder auf weiteres `!einwerfen` warten"
                )
            else:
                if len(self.game.einwerfen.spieler_liste) > 1:
                    anfaenger = self.game.einwerfen.spieler_liste[0]
                    out_str += f"\n{self.name_to_member(anfaenger.name).mention} hat mit einer {self.emoji_by_name(self._wuerfel_emoji_names[anfaenger.augen])} den niedrigsten Wurf und darf anfangen."
                    out_str += "\n`!wuerfeln` um das Spiel zu beginnen oder auf weiteres `!einwerfen` warten."
            await self.print_to_channel(channel, out_str)

        elif state_new == "stechen":
            spieler = self.spieler_by_name(
                spieler_name, self.game.einwerfen.spieler_liste
            )
            noch_stechen = [
                sp
                for sp in self.game.einwerfen.stecher_liste
                if sp not in self.game.einwerfen.gestochen_liste
            ]
            out_str = f"{message.author.mention} sticht mit einer {self.emoji_by_name(self._wuerfel_emoji_names[spieler.augen])}."
            if len(noch_stechen) > 1:
                out_str += "\n" + (
                    ", ".join(
                        [self.name_to_member(pl.name).mention for pl in noch_stechen]
                    )
                    + f" müssen `!stechen`."
                )
            elif len(noch_stechen) == 1 and len(self.game.einwerfen.stecher_liste) > 1:
                out_str += f"\n {self.name_to_member(noch_stechen[0].name).mention} muss noch `!stechen`."
            else:
                anfaenger = self.game.einwerfen.spieler_liste[0]
                out_str += f"\n{self.name_to_member(anfaenger.name).mention} hat mit einer {self.emoji_by_name(self._wuerfel_emoji_names[anfaenger.augen])} den niedrigsten Wurf und darf anfangen."
                out_str += "\n`!wuerfeln` um das Spiel zu beginnen."

            await self.print_to_channel(channel, out_str)

        elif state_new == "wuerfeln":
            # get old state information
            state_old = game_old.state.leaf_state.name
            stack_list_old = list(game_old.state_stack.deque)
            stack_names_old = [st.name for st in stack_list_old]
            num_halbzeit_old = stack_names_old.count("Halbzeit") + 1
            halbzeit_old = getattr(
                game_old, self._halbzeit_state_names[num_halbzeit_old]
            )

            stack_list = list(self.game.state_stack.deque)
            stack_names = [st.name for st in stack_list]
            num_halbzeit = stack_names.count("Halbzeit") + 1

            halbzeit = getattr(self.game, self._halbzeit_state_names[num_halbzeit])

            new_spieler_liste = halbzeit.spieler_liste
            spieler = self.spieler_by_name(spieler_name, new_spieler_liste)

            # Halbzeit beginnt
            if (
                state_old in ["einwerfen", "stechen"]
                or num_halbzeit_old != num_halbzeit
            ):
                old_spieler_liste = new_spieler_liste[:1] + new_spieler_liste[1:]
                # Erster wurf nach einwerfen. Print reihenfolge
                out_str0 = f"Halbzeit {num_halbzeit} beginnt. Die Reihenfolge ist:\n"
                out_str0 += ", ".join(
                    [self.name_to_member(pl.name).mention for pl in old_spieler_liste]
                )
                await self.print_to_channel(channel, out_str0)

            hoch, tief = halbzeit.rdm.hoch_und_tief()
            # handle wuerfeln
            augen = spieler.augen
            wurf_emoji = self.wurf_to_emoji(augen)
            # besonderer wurf?
            augen_name = str(wurf.welcher_wurf(augen))

            if hoch == tief:
                # erster Wurf der Runde
                out_str = f"{message.author.mention} legt vor: "
                is_erster_wurf = True
            else:
                out_str = f"{message.author.mention} wirft "
                is_erster_wurf = False
            out_str += wurf_emoji + "."

            reicht = tief.spieler.name != spieler.name
            comment_choices = [" "]
            if "Gemuese" in augen_name:
                if augen[0] < 5:
                    comment_choices = [
                        "Gar nicht mal so gut...",
                        "Schlechtes Gemüse...",
                        "Das kannst du besser!",
                        "Wow.",
                    ]
                    reicht_choices = {
                        "reicht": [" Aber reicht sogar.",],
                        "reichtnicht": [" Und reicht nicht mal.",],
                    }

                elif augen[0] == 5:
                    comment_choices = [
                        "Solides Gemüse.",
                        "Das kann man noch schlagen.",
                        "Ausbaufähig...",
                    ]
                    reicht_choices = {
                        "reicht": [" Und reicht sogar.",],
                        "reichtnicht": [" Aber reicht gar nicht.",],
                    }

                elif augen[0] == 6:
                    comment_choices = [
                        "Hohes Gemüse.",
                        "Nicht schlecht!",
                    ]
                    reicht_choices = {
                        "reicht": [" Und reicht sogar.",],
                        "reichtnicht": [" Aber reicht gar nicht.",],
                    }

            elif "General" in augen_name:
                comment_choices = ["Kann man liegen lassen.", "General."]
                reicht_choices = {
                    "reicht": [" Reicht ja.",],
                    "reichtnicht": [" Aber reicht gar nicht.",],
                }

            elif "Schock" in augen_name:
                if "out" in augen_name:
                    comment_choices = [
                        "Nice.",
                        "Random Schock Out",
                        "Würde ich liegen lassen",
                    ]
                    reicht_choices = {
                        "reicht": [" Reicht auch.",],
                        "reichtnicht": [" Aber reicht ja nicht mal.",],
                    }
                else:
                    comment_choices = ["Schöner Schock."]
                    reicht_choices = {
                        "reicht": [" Reicht auch.",],
                        "reichtnicht": [" Aber reicht gar nicht.",],
                    }

            elif "Herrenwurf" in augen_name:
                comment_choices = ["Herrenwurf. Verliert nicht."]
                reicht_choices = {
                    "reicht": [" Und reicht sogar.",],
                    "reichtnicht": [" ...aber vielleicht schon.",],
                }

            elif "Jule" in augen_name:
                comment_choices = ["Schöne Jule."]
                reicht_choices = {
                    "reicht": [" Und sie reicht.",],
                    "reichtnicht": [" Aber reicht leider nicht.",],
                }

            out_str += "\n" + random.choice(comment_choices)
            if not is_erster_wurf:
                if reicht:
                    out_str += random.choice(reicht_choices["reicht"])
                else:
                    out_str += random.choice(reicht_choices["reichtnicht"])

            if spieler.anzahl_wuerfe == 0 or game_cmd == "weiter":
                naechster = self.name_to_member(halbzeit.aktiver_spieler.name)
                out_str += f"\nAls nächstes ist {naechster.mention} an "
                out_str += f"der Reihe. Bitte `!wuerfeln`\n"
                # TODO figure out how to get high/low aus dem letzten durchgang
                hoch_spieler = hoch.spieler
                tief_spieler = tief.spieler
                hoch_augen = hoch.spieler.augen
                tief_augen = tief.spieler.augen
                out_str += f"Hoch ist {self.name_to_member(hoch_spieler.name).mention} "
                out_str += f"mit: {self.wurf_to_emoji(hoch_augen)}\n"
                out_str += f"Tief ist {self.name_to_member(tief_spieler.name).mention} "
                out_str += f"mit: {self.wurf_to_emoji(tief_augen)}\n"
            await self.print_to_channel(channel, out_str)



from schocken.wurf import welcher_wurf, SonderWurf, Gemuese


def test_aus_der_hand():
    Jule = SonderWurf.Jule
    assert welcher_wurf(Jule._value_, aus_der_hand=True) is Jule
    assert isinstance(welcher_wurf(Jule._value_, aus_der_hand=False), Gemuese)



from schocken import wurf


def test_wurf_prioritaet(N=100):
    wuerfe = [
        wurf.Schock.out,
        wurf.SonderWurf.Jule,
        wurf.Schock.sechs,
        wurf.Schock.fuenf,
        wurf.Schock.vier,
        wurf.Schock.drei,
        wurf.Schock.doof,
        wurf.General.sechser,
        wurf.General.fuenfer,
        wurf.General.vierer,
        wurf.General.dreier,
        wurf.General.zweier,
        wurf.Straße.vierer,
        wurf.Straße.dreier,
        wurf.Straße.zweier,
        wurf.Straße.einser,
    ]
    wurf_prio_soll = [wurf.prioritaet(W) for W in wuerfe]

    wurf_prio_ist = sorted(wurf_prio_soll)
    assert wurf_prio_soll == wurf_prio_ist



class Spieler:
    def __init__(self, name):
        self.name = name
        self.deckel = 0
        self.beiseite_gelegt = False
        self.einsen = 0
        self.augen = (None, None, None)
        self.anzahl_wuerfe = 0

    def __str__(self):
        return f"Spieler {self.name}\n\tDeckel: {self.deckel}\n\tAugen: {self.augen}\n"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, Spieler):
            return self.name == other.name
        return NotImplemented



import typing as T
import pysm
import threading
from . import events, wuerfel
from .deckel_management import RundenDeckelManagement, SpielzeitStatus
from .exceptions import (
    DuHastMistGebaut,
    FalscheAktion,
    FalscherSpieler,
    ZuOftGeworfen,
    NochNichtGeworfen,
    RundeVorbei,
    LustWurf,
)
from .spieler import Spieler


class Einwerfen(pysm.StateMachine):
    def __init__(self):
        super().__init__("Einwerfen")
        self.init_sm()
        self.spieler_liste = []
        self.stecher_count = 0
        self.stecher_liste = []
        self.gestochen_liste = []

    def init_sm(self):
        idle = pysm.State("einwerfen")
        stechen = pysm.State("stechen")

        self.add_state(idle, initial=True)
        self.add_state(stechen)

        idle.handlers = {
            "einwerfen": self.einwurf_handler,
            "wuerfeln": self.wuerfeln_handler,
            "exit": self.spieler_liste_fixieren,
        }

        stechen.handlers = {
            "stechen": self.stechen_handler,
            "einwerfen": self.raise_falsche_aktion,
            "wuerfeln": self.wuerfeln_handler,
            "exit": self.spieler_liste_fixieren,
        }

        self.add_transition(
            idle,
            stechen,
            events=["stechen"],
            condition=self.stechen_possible,
            after=self.stechen_handler,
        )

        self.initialize()

    def einwurf_handler(self, state, event):
        """Called when event "einwerfen" is dispatched"""
        spieler_name = event.cargo["spieler_name"]
        if spieler_name in [sp.name for sp in self.spieler_liste]:
            raise FalscherSpieler

        spieler = Spieler(spieler_name)
        einwurf = wuerfel.werfen(1)[0]

        spieler.augen = einwurf
        self.spieler_liste.append(spieler)

        roll_list = [sp.augen for sp in self.spieler_liste]

        self.stecher_liste = [
            sp for sp in self.spieler_liste if sp.augen == min(roll_list)
        ]
        self.stecher_count = len(self.stecher_liste)

    def stechen_handler(self, state, event):
        spieler_name = event.cargo["spieler_name"]
        if len(self.gestochen_liste) == 0:
            self._init_stecher_count = len(self.stecher_liste)

        # check if already gestochen
        if spieler_name in [pl.name for pl in self.gestochen_liste]:
            raise FalscherSpieler

        # check if eligible
        if spieler_name not in [st.name for st in self.stecher_liste]:
            raise FalscherSpieler

        stich = wuerfel.werfen(1)[0]
        stecher = [sp for sp in self.spieler_liste if sp.name == spieler_name][0]
        stecher.augen = stich

        self.gestochen_liste.append(stecher)
        # if all stiche done, determine starting player or stech again
        if len(self.gestochen_liste) == self._init_stecher_count:
            stich_list = [st.augen for st in self.gestochen_liste]
            self.stecher_liste = [
                sp for sp in self.gestochen_liste if sp.augen == min(stich_list)
            ]
            self.gestochen_liste = []
            # sort stecher by stich
        elif len(self.gestochen_liste) < self._init_stecher_count:
            pass

        self.stecher_count = len(self.stecher_liste)

    def wuerfeln_handler(self, state, event):
        spieler_name = event.cargo["spieler_name"]
        if not self.wuerfeln_possible():
            raise FalscheAktion
        elif spieler_name != self.stecher_liste[0].name:
            raise FalscherSpieler(f"Nur {self.stecher_liste[0].name} darf anfangen!")

    def raise_falsche_aktion(self, state, event):
        raise FalscheAktion

    def stechen_possible(self, state, event):
        if len(self.spieler_liste) > 1 and self.stecher_count > 1:
            return True
        else:
            raise FalscheAktion("Stechen geht gerade nicht")

    def wuerfeln_possible(self):
        return len(self.spieler_liste) > 1 and self.stecher_count <= 1

    def sortierte_spieler_liste(self):
        try:
            return self.spieler_liste_fixiert
        except AttributeError as err:
            raise DuHastMistGebaut("Einwerfen war noch nicht vorbei!") from err

    def spieler_liste_fixieren(self, state, event):
        spieler_liste = self.spieler_liste
        if self.state.name == "stechen":
            # rotate spieler_liste according to lowest stecher
            spieler = self.stecher_liste[0]
            idx = self.spieler_liste.index(spieler)
            spieler_liste = spieler_liste[idx:] + spieler_liste[:idx]
        else:
            # rotate spieler_liste such that lowest roll is first element
            roll_list = [sp.augen for sp in spieler_liste]
            min_roll = min(roll_list)
            min_index = roll_list.index(min_roll)
            spieler_liste = spieler_liste[min_index:] + spieler_liste[:min_index]
        self.spieler_liste_fixiert = spieler_liste


class Halbzeit(pysm.StateMachine):
    def __init__(self):
        super().__init__("Halbzeit")
        self.verlierende = None
        self.spielzeit_status = None
        self.rdm = None
        self.letzter_wurf = (None, None, None)

        self.handlers = {"enter": self.enter, "after": self.after}

        wuerfeln = pysm.State("wuerfeln")
        wuerfeln.handlers = {
            "wuerfeln": self.wuerfeln_handler,
            "beiseite legen": self.beiseite_legen_handler,
            "weiter": self.naechster_spieler_handler,
            "umdrehen": self.sechsen_handler,
        }
        self.add_state(wuerfeln, initial=True)

        self.initialize()

    def enter(self, state, event):
        num_halbzeit = [
            x for x in self.parent.state_stack.deque if x.name == "Halbzeit"
        ]
        if len(num_halbzeit) > 1:
            verlierende_1 = self.root_machine.halbzeit_erste.verlierende
            verlierende_2 = self.root_machine.halbzeit_zweite.verlierende
            spieler_liste = [verlierende_2, verlierende_1]
        else:
            vorheriger_state = self.root_machine.state_stack.peek()
            spieler_liste = vorheriger_state.sortierte_spieler_liste()
            self.initiale_spieler = spieler_liste.copy()

        self._spielerinnen_unique = set(s.name for s in spieler_liste)
        for s in spieler_liste:
            s.deckel = 0
        self.spielzeit_status = SpielzeitStatus(15, spieler_liste)
        self.rdm = RundenDeckelManagement(self.spielzeit_status)

    def after(self, state, event):
        if len(self._spielerinnen_unique) == 1:
            self.root_machine.dispatch(pysm.Event(events.FERTIG_HALBZEIT))

    @property
    def spieler_liste(self) -> T.List[Spieler]:
        return self.spielzeit_status.spieler

    @property
    def aktiver_spieler(self) -> T.Optional[Spieler]:
        return self.rdm.aktiver_spieler if self.rdm else None

    def sortierte_spieler_liste(self):
        if not self.verlierende:
            raise DuHastMistGebaut("Es gibt noch keinen definierten Startspieler")
        for idx, spieler in enumerate(self.initiale_spieler):
            if spieler.name == self.verlierende.name:
                break
        else:
            raise DuHastMistGebaut(
                f"Der/die Verlierende `{self.verlierende.name}`)` spielt gar nicht "
                f"mit! Mitspielende: {self.initiale_spieler}"
            )
        sotiert = self.initiale_spieler[idx:] + self.initiale_spieler[:idx]
        return sotiert

    def wuerfeln_handler(self, state, event):
        lust_wurf_geworfen = False
        akt_spieler = self.aktiver_spieler
        spieler_name = event.cargo["spieler_name"]

        if spieler_name != akt_spieler.name:
            raise FalscherSpieler(
                f"{spieler_name} wollte würfeln, {akt_spieler.name} ist aber dran!"
            )

        if akt_spieler.anzahl_wuerfe < self.rdm.num_maximale_wuerfe:
            # check if ones were put aside
            if akt_spieler.einsen > 0:
                wurf = wuerfel.werfen(3 - akt_spieler.einsen)
                akt_spieler.augen = (1,) * akt_spieler.einsen + wurf
                akt_spieler.anzahl_wuerfe += 1
                akt_spieler.beiseite_gelegt = False
                aus_der_hand = False
            else:
                akt_spieler.augen = wuerfel.werfen(3)
                akt_spieler.anzahl_wuerfe += 1
                aus_der_hand = True
            letzter_wurf = self.rdm.wurf(spieler_name, akt_spieler.augen, aus_der_hand)
            if self.rdm.ist_lust_wurf(letzter_wurf):
                self.rdm.strafdeckel_verteilen(letzter_wurf.spieler)
                lust_wurf_geworfen = True
        else:
            # watch for semantics
            num_wurf = self.rdm.num_maximale_wuerfe
            plural_switch = "Wurf ist" if num_wurf == 1 else "Würfe sind"
            zahl_zu_wort = {1: "ein", 2: "zwei", 3: "drei"}
            meldung = (
                f"Maximal {zahl_zu_wort[num_wurf]} {plural_switch} erlaubt, "
                f"{akt_spieler.name}!"
            )
            raise ZuOftGeworfen(meldung)

        if akt_spieler.anzahl_wuerfe == self.rdm.num_maximale_wuerfe:
            akt_spieler.anzahl_wuerfe = 0
            akt_spieler.einsen = 0
            try:
                self.rdm.weiter()
            except RundeVorbei:
                self.spielzeit_status = self.rdm.deckel_verteilen_restliche_spieler()
                if self.beendet():
                    self.verlierende = self.spielzeit_status.spieler[0]
                    self.root_machine.dispatch(pysm.Event(events.FERTIG_HALBZEIT))
                else:
                    self.rdm = RundenDeckelManagement(self.spielzeit_status)

        if lust_wurf_geworfen:
            raise LustWurf(letzter_wurf)

    def beiseite_legen_handler(self, state, event):
        akt_spieler = self.aktiver_spieler
        spieler_name = event.cargo["spieler_name"]

        if spieler_name != akt_spieler.name:
            raise FalscherSpieler(
                f"Das kannst du zur Zeit nicht tun, {spieler_name}. "
                f"{akt_spieler.name} ist an der Reihe!"
            )

        if akt_spieler.beiseite_gelegt:
            raise FalscheAktion(f"Du hast bereits beiseite gelegt!")
        elif not akt_spieler.beiseite_gelegt and 1 in akt_spieler.augen:
            akt_spieler.einsen += akt_spieler.augen.count(1) - akt_spieler.einsen
            akt_spieler.beiseite_gelegt = True
        else:
            raise FalscheAktion(
                f"Du hast keine Einsen gewürfelt die du zur Seite legen kannst!"
            )

    def naechster_spieler_handler(self, state, event):
        akt_spieler = self.aktiver_spieler
        spieler_name = event.cargo["spieler_name"]

        if spieler_name != akt_spieler.name:
            raise FalscherSpieler(
                f"Das kannst du zur Zeit nicht tun, {spieler_name}. "
                f"{akt_spieler.name} ist an der Reihe!"
            )

        if akt_spieler.anzahl_wuerfe == 0:
            raise NochNichtGeworfen("Es muss mindestens ein Mal gewürfelt werden!")
        else:
            akt_spieler.anzahl_wuerfe = 0
            self.rdm.weiter()

    def sechsen_handler(self, state, event):
        akt_spieler = self.aktiver_spieler
        spieler_name = event.cargo["spieler_name"]

        if spieler_name != akt_spieler.name:
            raise FalscherSpieler(
                f"Das kannst du zur Zeit nicht tun, {spieler_name}. "
                f"{akt_spieler.name} ist an der Reihe!"
            )

        if 6 not in akt_spieler.augen:
            raise FalscheAktion("Du hast keine Sechsen zum Umdrehen!")
        elif akt_spieler.augen.count(6) == 1:
            raise FalscheAktion("Du hast nur eine Sechs gewürfelt!")
        elif akt_spieler.beiseite_gelegt:
            raise FalscheAktion(
                f"Du hast bereits deine Sechsen zu einer Eins umgewandelt!"
            )
        else:
            akt_spieler.einsen += 1
            akt_spieler.beiseite_gelegt = True

    def beendet(self):
        return len(self.spieler_liste) == 1


class SchockenSpiel(pysm.StateMachine):
    def __init__(self):
        super().__init__("SchockenSpiel")
        self.einwerfen = Einwerfen()
        self.halbzeit_erste = Halbzeit()
        self.halbzeit_zweite = Halbzeit()
        self.finale = Halbzeit()
        anstoßen = pysm.State("anstoßen!")

        # add states to machine
        self.add_state(self.einwerfen, initial=True)
        self.add_state(self.halbzeit_erste)
        self.add_state(self.halbzeit_zweite)
        self.add_state(self.finale)
        self.add_state(anstoßen)

        self.add_transition(
            self.einwerfen,
            self.halbzeit_erste,
            events=[events.WÜRFELN],
            after=self.halbzeit_erste.wuerfeln_handler,
        )

        self.add_transition(
            self.halbzeit_erste, self.halbzeit_zweite, events=[events.FERTIG_HALBZEIT],
        )
        self.add_transition(
            self.halbzeit_zweite,
            self.finale,
            events=[events.FERTIG_HALBZEIT],
            after=self.finale.after,
        )
        self.add_transition(
            self.finale, anstoßen, events=[events.FERTIG_HALBZEIT], after=self.anstoßen,
        )
        self.initialize()

    def command_to_event(self, spieler_name, command):
        # please stick to the convention that event identifiers are the same
        # as the command strings
        if command == "einwerfen":
            event = pysm.Event("einwerfen", spieler_name=spieler_name)
        elif command == "wuerfeln":
            event = pysm.Event("wuerfeln", spieler_name=spieler_name)
        elif command == "stechen":
            event = pysm.Event("stechen", spieler_name=spieler_name)
        elif command == "weiter":
            event = pysm.Event("weiter", spieler_name=spieler_name)
        elif command == "beiseite legen":
            event = pysm.Event("beiseite legen", spieler_name=spieler_name)
        elif command == "umdrehen":
            event = pysm.Event("umdrehen", spieler_name=spieler_name)
        else:
            raise FalscheAktion
        self.dispatch(event)

    def anstoßen(self, state, event):
        print("PROST!")



class FalscheAktion(ValueError):
    pass


class FalscherServer(ValueError):
    pass


class FalscherSpielBefehl(ValueError):
    pass


class SpielLaeuft(ValueError):
    pass


class SpielLaeuftNicht(ValueError):
    pass


class ZuWenigSpieler(ValueError):
    pass


class FalscherSpieler(ValueError):
    pass


class NochNichtGeworfen(ValueError):
    pass


class ZuOftGeworfen(ValueError):
    pass


class UnbekannterSpieler(ValueError):
    pass


class KeineWuerfeVorhanden(ValueError):
    pass


class DuHastMistGebaut(RuntimeError):
    pass


class RundeVorbei(ValueError):
    pass


class PermissionError(ValueError):
    pass


class LustWurf(RuntimeError):
    def __init__(self, wurf):
        self.wurf = wurf



from offline_test_helpers import (
    MockBot,
    MockClient,
    MockMember,
    MockMessage,
)

# from schocken.exceptions import SpielLaeuft
from schocken import wuerfel
import pytest


@pytest.fixture
def member(n=4):
    return [MockMember(f"spieler_{i+1}") for i in range(n)]


@pytest.fixture
def bot():
    client = MockClient()
    bot = MockBot(client)
    return bot


pytestmark = pytest.mark.asyncio


async def test_start_game(capsys, member, bot):
    assert not bot.game_running
    await bot.parse_input(MockMessage(member[0], "!schocken"))
    assert bot.game_running
    out, err = capsys.readouterr()
    # richtiger spieler:
    assert "spieler_1" in out
    # start message sollte !einwerfen drin haben:
    assert "!einwerfen" in out
    # anderer spieler will auch starten
    await bot.parse_input(MockMessage(member[1], "!schocken"))
    out, err = capsys.readouterr()
    assert "läuft" in out


async def test_einwerfen(member, bot):
    # spiel starten mit fixture ist schwierig wegen der coroutines
    # spiel manuell starten:
    await bot.parse_input(MockMessage(member[0], "!schocken"))
    assert bot.game_running

    print(bot.game.state)
    wuerfel.werfen = lambda n: (2,) * n
    await bot.parse_input(MockMessage(member[0], "!einwerfen"))

    wuerfel.werfen = lambda n: (2,) * n
    await bot.parse_input(MockMessage(member[0], "!einwerfen"))

    # out, _ = capsys.readouterr()
    # print(out)


async def main():
    client = MockClient()
    bot = MockBot(client)

    spieler_1 = MockMember("spieler_1")
    spieler_2 = MockMember("spieler_2")
    spieler_3 = MockMember("spieler_3")
    spieler_4 = MockMember("spieler_4")

    # spieler_1 starts game
    await bot.parse_input(MockMessage(spieler_1, "!schocken"))

    wuerfel.werfen = lambda n: (2,) * n
    await bot.parse_input(MockMessage(spieler_1, "!einwerfen"))
    # einwerfen, stechen testen
    wuerfel.werfen = lambda n: (1,) * n
    await bot.parse_input(MockMessage(spieler_2, "!einwerfen"))
    await bot.parse_input(MockMessage(spieler_3, "!einwerfen"))

    wuerfel.werfen = lambda n: (3,) * n
    await bot.parse_input(MockMessage(spieler_2, "!stechen"))

    wuerfel.werfen = lambda n: (3,) * n
    await bot.parse_input(MockMessage(spieler_3, "!stechen"))

    wuerfel.werfen = lambda n: (2,) * n
    await bot.parse_input(MockMessage(spieler_3, "!stechen"))

    wuerfel.werfen = lambda n: (3,) * n
    await bot.parse_input(MockMessage(spieler_2, "!stechen"))

    # Einwerfen vorbei, jetzt würfeln.
    await bot.parse_input(MockMessage(spieler_3, "!wuerfeln"))



from random import randint


class ZuVieleWuerfel(Exception):
    pass


def werfen(anzahl):
    if anzahl <= 3:
        return tuple(sorted([randint(1, 6) for _ in range(anzahl)], reverse=True))
    else:
        raise ZuVieleWuerfel(
            f"Es dürfen maximal 3 Würfel geworfen werden! Nicht {anzahl}."
        )



import pytest

from schocken import events
from schocken.spiel import SchockenSpiel
from schocken.spieler import Spieler
from schocken.exceptions import FalscheAktion, FalscherSpieler

from schocken import wuerfel


@pytest.fixture
def spieler(n=4):
    # testspieler
    return [Spieler(f"spieler_{i+1}") for i in range(n)]


def test_einwerfen_trotz_stechen(spieler):
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"

    # FALL: Zwei Spieler werfen dasselbe ein, einer versucht zu starten
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    runde.command_to_event(spieler[1].name, "einwerfen")

    with pytest.raises(FalscheAktion):
        runde.command_to_event(spieler[1].name, events.WÜRFELN)


@pytest.fixture
def drei_spieler_eingeworfen_spieler_zwei_muss_werfen(spieler):
    # FALL: spieler 2 würfelt eindeutig das niedrigste
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[1].name, "einwerfen")
    wuerfel.werfen = lambda n: (3,) * n
    runde.command_to_event(spieler[2].name, "einwerfen")
    return runde


def test_falscher_spieler(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen
    # Richtige exception, wenn jemand anders anfangen will?
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, events.WÜRFELN)


def test_falsche_aktion(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen
    # Richtige exception, wenn jemand stechen will.
    with pytest.raises(FalscheAktion):
        runde.command_to_event(spieler[1].name, "stechen")


def test_korrekter_start(spieler, drei_spieler_eingeworfen_spieler_zwei_muss_werfen):
    runde = drei_spieler_eingeworfen_spieler_zwei_muss_werfen
    # momentaner state ist einwerfen
    assert runde.leaf_state.name == "einwerfen"

    # spieler[1] wuerfelt, danach muss einwerfen fertig sein
    runde.command_to_event(spieler[1].name, events.WÜRFELN)
    assert runde.leaf_state.name == events.WÜRFELN


@pytest.fixture
def vierfach_stechen(spieler):
    runde = SchockenSpiel()
    assert runde.leaf_state.name == "einwerfen"
    wuerfel.werfen = lambda n: (1,) * n
    for S in spieler:
        runde.command_to_event(S.name, "einwerfen")
    return runde


def test_mehrfaches_stechen(spieler, vierfach_stechen):
    # FALL mehrfaches stechen, erst 4, dann 3, dann 2 spieler, dann beginn
    runde = vierfach_stechen

    # alle stecher in der liste?
    stecher = [st.name for st in runde.einwerfen.stecher_liste]
    assert stecher == [S.name for S in spieler]

    # neues stechen, 3 spieler werfen 1, spieler_4 wirft 2
    runde.command_to_event(spieler[0].name, "stechen")
    assert runde.leaf_state.name == "stechen"
    runde.command_to_event(spieler[1].name, "stechen")
    runde.command_to_event(spieler[2].name, "stechen")
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[3].name, "stechen")
    # spieler 4 versucht unerlaubterweise zu stechen
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[3].name, "stechen")

    # die anderen spieler stechen
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[0].name, "stechen")
    runde.command_to_event(spieler[1].name, "stechen")
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[2].name, "stechen")
    # nur noch 1 und 2 in der stecherliste
    stecher = [st.name for st in runde.einwerfen.stecher_liste]
    assert stecher == [S.name for S in spieler[:2]]

    # 1 und 2 stechen
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[0].name, "stechen")
    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[1].name, "stechen")

    # spieler 1 versucht unerlaubterweise anzufangen
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, events.WÜRFELN)

    # spieler 2 faengt an
    runde.command_to_event(spieler[1].name, events.WÜRFELN)
    assert runde.leaf_state.name == events.WÜRFELN
    assert runde.einwerfen.sortierte_spieler_liste()[0].name == spieler[1].name


def test_einwerfen_waehrend_stechen(spieler):
    # FALL einwerfen wenn stechen im gange ist
    runde = SchockenSpiel()

    wuerfel.werfen = lambda n: (1,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    runde.command_to_event(spieler[1].name, "einwerfen")

    # spieler 2 will nochmal einwerfen
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, "einwerfen")

    runde.command_to_event(spieler[0].name, "stechen")
    # weiterer spieler will noch einwerfen, obwohl gestochen wird
    with pytest.raises(FalscheAktion):
        runde.command_to_event(spieler[2].name, "einwerfen")


def test_stiche_höher_einwurf_augen(spieler):
    # FALL Stiche sind größer als nächstkleinere Einwurf Augenzahlen
    runde = SchockenSpiel()

    assert runde.leaf_state.name == "einwerfen"
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[0].name, "einwerfen")
    wuerfel.werfen = lambda n: (3,) * n
    runde.command_to_event(spieler[1].name, "einwerfen")
    wuerfel.werfen = lambda n: (2,) * n
    runde.command_to_event(spieler[2].name, "einwerfen")

    wuerfel.werfen = lambda n: (6,) * n
    runde.command_to_event(spieler[0].name, "stechen")

    wuerfel.werfen = lambda n: (5,) * n
    runde.command_to_event(spieler[2].name, "stechen")

    # spieler 3 muss anfangen zu würfeln
    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[0].name, events.WÜRFELN)

    with pytest.raises(FalscherSpieler):
        runde.command_to_event(spieler[1].name, events.WÜRFELN)



from schocken.bot import SchockenBot


class LogoutException(ValueError):
    pass


class MockGuild:
    def __init__(self, name):
        self.name = name
        self.emojis = []
        self.members = [
            MockMember("spieler_1"),
            MockMember("spieler_2"),
            MockMember("spieler_3"),
            MockMember("spieler_4"),
        ]


class MockClient:
    def __init__(self):
        self.guilds = [MockGuild("Café A")]

    def logout(self):
        raise LogoutException("client.logout() called")


class MockMember:
    def __init__(self, name, roles=["Schichtler/in"]):
        self.name = name
        self.mention = f"MENTION:{name}"
        self.roles = roles


class MockChannel:
    def __init__(self, name):
        self.name = name

    def send(self, message):
        return f"MSG to channel {self.name}:\n {message}"


class MockMessage:
    def __init__(self, author, content, channel_name="schocktresen"):
        self.content = content
        self.author = author
        self.channel = MockChannel(channel_name)


class MockBot(SchockenBot):
    # override methods that need a server connection
    def emoji_by_name(self, name):
        return f"EMOJI:{name}"

    def name_to_member(self, name):
        return MockMember(name)

    async def print_to_channel(self, channel, text):
        msg = channel.send(text)
        print(msg + "\n" + "-" * 72)



import toml
from setuptools import setup, find_namespace_packages

package_dir = "src"
pyproject = toml.load("pyproject.toml")

# install dev using pipenv:
# pipenv install -e. ".[dev]"

setup(
    **pyproject["project"],
    install_requires=pyproject["dependencies"].keys(),
    extras_require={"dev": pyproject["dev-dependencies"].keys()},
    package_dir={"": package_dir},
    packages=find_namespace_packages(where=package_dir),
    entry_points={"console_scripts": ["schocken = schocken.__main__:run_bot",],},
)



import typing as T
from copy import deepcopy

from .wurf import welcher_wurf, Wurf, prioritaet, Schock
from .spieler import Spieler
from .exceptions import (
    ZuWenigSpieler,
    NochNichtGeworfen,
    ZuOftGeworfen,
    FalscherSpieler,
    KeineWuerfeVorhanden,
    UnbekannterSpieler,
    RundeVorbei,
)

NUM_MAX_DECKEL = 15


class WurfEvaluierung(T.NamedTuple):
    prioritaet: float
    wurf_anzahl: int
    reihenfolge: int
    spieler: Spieler
    wurf: Wurf


class SpielzeitStatus(T.NamedTuple):
    deckel_in_topf: int
    spieler: T.List[Spieler]


class RundenDeckelManagement:
    def __init__(self, runden_status: SpielzeitStatus):
        if len(runden_status.spieler) < 2:
            raise ZuWenigSpieler

        self._waren_anfangs_deckel_im_topf = runden_status.deckel_in_topf > 0
        self._zahl_deckel_im_topf = runden_status.deckel_in_topf
        self._spieler = runden_status.spieler
        self._spieler_namen = [S.name for S in self._spieler]
        self._aktueller_spieler_idx = 0
        self._wuerfe = {S.name: [] for S in self._spieler}

    def weiter(self) -> int:
        aktueller_spieler = self._spieler_namen[self._aktueller_spieler_idx]
        if not self._wuerfe[aktueller_spieler]:
            raise NochNichtGeworfen
        if self._aktueller_spieler_idx + 1 >= len(self._spieler):
            raise RundeVorbei("Runde ist bereits vorbei!")
        self._aktueller_spieler_idx += 1
        return self._aktueller_spieler_idx

    def deckel_verteilen_restliche_spieler(self) -> SpielzeitStatus:
        hoch, tief = self.hoch_und_tief()
        self._deckel_verteilen(hoch, tief)
        restliche_spieler = self._restliche_spielerinnen_bestimmen(tief.spieler)
        return SpielzeitStatus(self._zahl_deckel_im_topf, restliche_spieler)

    def _deckel_verteilen(self, hoch: WurfEvaluierung, tief: WurfEvaluierung):
        if hoch.wurf is Schock.out:
            self._schockout_verteilen(tief)
        elif self._waren_anfangs_deckel_im_topf:
            self._deckel_im_topf_verteilen_hoch_tief(hoch, tief)
        else:
            self._deckel_von_hoch_an_tief_verteilen(hoch, tief)

    def _deckel_im_topf_verteilen_hoch_tief(
        self, hoch: WurfEvaluierung, tief: WurfEvaluierung
    ):
        anzahl_deckel = hoch.wurf.deckel_wert
        self._deckel_im_topf_verteilen(tief.spieler, anzahl_deckel)

    def _deckel_im_topf_verteilen(self, spieler: Spieler, anzahl_deckel: int):
        anzahl_deckel = min(self._zahl_deckel_im_topf, anzahl_deckel)
        self._zahl_deckel_im_topf -= anzahl_deckel
        spieler.deckel += anzahl_deckel

    def _deckel_von_hoch_an_tief_verteilen(
        self, hoch: WurfEvaluierung, tief: WurfEvaluierung
    ):
        self._deckel_von_spieler_an_spieler(
            geber=hoch.spieler,
            empfaenger=tief.spieler,
            anzahl_deckel=hoch.wurf.deckel_wert,
        )

    def _deckel_von_spieler_an_spieler(
        self, geber: Spieler, empfaenger: Spieler, anzahl_deckel: int
    ):
        anzahl_deckel = min(geber.deckel, anzahl_deckel)
        geber.deckel -= anzahl_deckel
        empfaenger.deckel += anzahl_deckel

    def _schockout_verteilen(self, tief: WurfEvaluierung):
        zahl_verteile_deckel = sum(s.deckel for s in self._spieler)
        zahl_alle_deckel = self._zahl_deckel_im_topf + zahl_verteile_deckel
        self._zahl_deckel_im_topf = 0
        for s in self._spieler:
            s.deckel = 0
        tief.spieler.deckel = zahl_alle_deckel

    def _restliche_spielerinnen_bestimmen(self, verlierer: Spieler) -> T.List[Spieler]:
        spielerinnen = [s for s in self._spieler if self._noch_im_spiel(s)]
        start_idx = spielerinnen.index(verlierer)
        spielerinnen = spielerinnen[start_idx:] + spielerinnen[:start_idx]
        return spielerinnen

    def _noch_im_spiel(self, spielerin):
        return self._zahl_deckel_im_topf or spielerin.deckel

    def wurf(
        self,
        spieler_name: str,
        augen: T.Tuple[int, int, int],
        aus_der_hand: bool = True,
    ) -> WurfEvaluierung:
        try:
            spieler_idx = self._spieler_namen.index(spieler_name)
        except ValueError as err:
            raise UnbekannterSpieler(spieler_name) from err

        if spieler_idx != self._aktueller_spieler_idx:
            aktueller_spieler = self._spieler_namen[self._aktueller_spieler_idx]
            raise FalscherSpieler(
                f"{spieler_name} hat geworfen, {aktueller_spieler} war aber dran!"
            )

        bestehende_wuerfe = self._wuerfe[spieler_name]
        if len(bestehende_wuerfe) >= self.num_maximale_wuerfe:
            raise ZuOftGeworfen()

        wurf = welcher_wurf(augen, aus_der_hand)
        bestehende_wuerfe.append(wurf)

        return WurfEvaluierung(
            prioritaet(wurf),
            wurf_anzahl=len(bestehende_wuerfe),
            reihenfolge=spieler_idx,
            spieler=self._spieler[spieler_idx],
            wurf=wurf,
        )

    def ist_lust_wurf(self, wurf: WurfEvaluierung):
        if wurf.wurf_anzahl <= 1 or wurf.reihenfolge == 0:
            return False
        simulation = deepcopy(self)
        verteilung = simulation.deckel_verteilen_restliche_spieler()
        waere_raus = wurf.spieler not in verteilung.spieler
        return waere_raus

    def strafdeckel_verteilen(self, bestrafte: Spieler):
        if self._waren_anfangs_deckel_im_topf:
            self._deckel_im_topf_verteilen(bestrafte, 1)
        else:  # spieler mit den meisten Deckeln gibt ab
            deckel = [s.deckel for s in self._spieler]
            meiste_deckel_index = deckel.index(max(deckel))
            geber = self._spieler[meiste_deckel_index]
            self._deckel_von_spieler_an_spieler(geber, bestrafte, 1)
        # NOTE: Auslegungssache. Falls der bestrafte keinen Deckel bekommen hat, zb wenn
        # es keinen Deckel mehr im Topf gab, und nächste Runde trotzdem spielen müsste,
        # dann muss hier entsprechendes `spieler.bestraft` flag gesetzt werden, dass
        # dann in `_restliche_spielerinnen_bestimmen()` abgefragt werden muss.

    def hoch_und_tief(self) -> T.Tuple[WurfEvaluierung, WurfEvaluierung]:
        erster_spieler = self._spieler_namen[0]
        if not self._wuerfe[erster_spieler]:
            raise KeineWuerfeVorhanden()

        evaluierungen = []
        for idx, spieler in enumerate(self._spieler):
            bestehende_wuerfe = self._wuerfe[spieler.name]
            if not bestehende_wuerfe:
                break
            letzter_wurf = bestehende_wuerfe[-1]
            evaluierung = WurfEvaluierung(
                prioritaet=prioritaet(letzter_wurf),
                wurf_anzahl=len(bestehende_wuerfe),
                reihenfolge=idx,
                spieler=spieler,
                wurf=letzter_wurf,
            )
            evaluierungen.append(evaluierung)
        # hoch: index 0, tief: index -1
        evaluierungen.sort(key=self._schockout_reihenfolge_fix)
        return evaluierungen[0], evaluierungen[-1]

    @staticmethod
    def _schockout_reihenfolge_fix(evaluierung: WurfEvaluierung):
        """Schocks werden in umgekehrter Reihenfolge gewertet"""
        if evaluierung.wurf is Schock.out:
            evaluierung = WurfEvaluierung(
                prioritaet=evaluierung.prioritaet,
                wurf_anzahl=evaluierung.wurf_anzahl,
                reihenfolge=-evaluierung.reihenfolge,  # NOTE: Reihenfolge umgekehrt
                spieler=evaluierung.spieler,
                wurf=evaluierung.wurf,
            )
        return evaluierung

    @property
    def num_maximale_wuerfe(self):
        if self._aktueller_spieler_idx == 0:
            return 3
        else:
            start_spieler_name = self._spieler_namen[0]
            wuerfe_start_spieler = self._wuerfe[start_spieler_name]
            num_wuerfe_start_spieler = len(wuerfe_start_spieler)
            return num_wuerfe_start_spieler

    @property
    def aktiver_spieler(self):
        return self._spieler[self._aktueller_spieler_idx]



import typing as T
from enum import Enum

_ZAHL = {1: "eins", 2: "zwei", 3: "drei", 4: "vier", 5: "fuenf", 6: "sechs"}


class Wurf(Enum):
    pass


class SonderWurf(Wurf):
    Jule = (4, 2, 1)
    Herrenwurf = (6, 5, 5)

    @property
    def deckel_wert(self):
        return {self.Jule: 7, self.Herrenwurf: 1}[self]


_zahlen_3_bis_6 = tuple(map(lambda i: _ZAHL[i], range(3, 7)))
_schock_namen = ("out", "doof") + _zahlen_3_bis_6
_schocks = [
    (schock_name, (schock_suffix, 1, 1))
    for schock_name, schock_suffix in zip(_schock_namen, range(1, 7))
]
Schock = Wurf("Schock", _schocks)
Schock.deckel_wert = property(
    lambda self: 15 if self is Schock.out else self._value_[0]
)

_generaele = [(f"{_ZAHL[i]}er", (i,) * 3) for i in range(2, 7)]
General = Wurf("General", _generaele)
General.deckel_wert = property(lambda self: 3)

_zahlen_1_bis_4 = tuple(map(lambda i: _ZAHL[i], range(1, 5)))
_straßen_namen = (f"{zahl}er" for zahl in _zahlen_1_bis_4)
_straßen = [(name, (i + 3, i + 2, i + 1)) for i, name in enumerate(_straßen_namen)]
Straße = Wurf("Straße", _straßen)
Straße.deckel_wert = property(lambda self: 2)


class Gemuese:
    def __init__(self, augen):
        self._value_ = augen

    def __repr__(self):
        return f"<{type(self).__name__}: {self._value_}>"

    @property
    def deckel_wert(self):
        return 1


def welcher_wurf(wuerfel_augen: T.Tuple[int, int, int], aus_der_hand: bool = True):
    wuerfel_augen = tuple(sorted(wuerfel_augen, reverse=True))
    wurf_klassen = [Schock]
    if aus_der_hand:
        wurf_klassen += [SonderWurf, General, Straße]
    for wurf_klasse in wurf_klassen:
        if wuerfel_augen in [wurf._value_ for wurf in wurf_klasse]:
            return wurf_klasse(wuerfel_augen)
    else:
        return Gemuese(wuerfel_augen)


def prioritaet(wurf):
    prio = None
    if isinstance(wurf, Gemuese) or wurf is SonderWurf.Herrenwurf:
        prio = wurf._value_[0] / 10 + wurf._value_[1] / 100 + wurf._value_[2] / 1000
    elif wurf is SonderWurf.Jule:
        prio = 4.07
    elif wurf in Schock:
        prio = 4 + wurf.deckel_wert / 100
    elif wurf in General:
        prio = 3 + wurf._value_[0] / 10
    elif wurf in Straße:
        prio = 2 + wurf._value_[0] / 10
    else:
        raise ValueError(f"Unhandled wurf {wurf}")
    return -prio
